import os
from uuid import uuid4

import pytest

from slash import Session
from slash.exceptions import CannotLoadTests
from slash.loader import Loader

from .utils.suite_writer import Suite


def test_total_num_tests(suite):
    suite.debug_info = False

    path = suite.commit()

    with Session() as s:
        Loader().get_runnables(path)
        assert s.get_total_num_tests() == len(suite)


def test_loader_sort_filenames(tmpdir):

    tests_dir = tmpdir.join(str(uuid4()))

    filenames = []

    for _ in range(10):
        filename = str(uuid4()).replace('-', '') + '.py'

        with tests_dir.join(filename).open('w', ensure=True) as f:
            f.write('def test_something():\n    pass')

        filenames.append(filename)

    with Session():
        runnables = Loader().get_runnables(str(tests_dir))

    assert [os.path.basename(runnable.__slash__.file_path) for runnable in runnables] == sorted(filenames)


def test_loader_skips_empty_dirs(tmpdir):
    tests_dir = tmpdir.join('tests')
    with tests_dir.join('.dir').join('test_something.py').open('w', ensure=True) as f:
        f.write('def test_something():\n    pass')

    with Session():
        runnables = Loader().get_runnables(str(tests_dir))

    assert runnables == []


def test_loader_warns_duplicate_test_funcs(tmpdir):
    tests_dir = tmpdir.join('tests')
    full_path = tests_dir.join('.dir').join('test_something.py')
    test_name = 'test_something'
    with full_path.open('w', ensure=True) as f:
        f.write('def {}():\n    assert True\n'.format(test_name))
        f.write('def {}():\n    assert True\n'.format(test_name))
    with Session() as session:
        Loader().get_runnables([str(full_path)])
        assert len(session.warnings) == 1
        assert 'Duplicate' in session.warnings.warnings[0].details['message']
        assert test_name in session.warnings.warnings[0].details['message']


def test_loader_warns_on_duplicate_fixtures(suite):
    fixture_name = 'fixture_name'
    fixture1 = suite.slashconf.add_fixture(name=fixture_name)
    fixture1.append_line('assert this == slash.context.fixture')
    fixture2 = suite.slashconf.add_fixture(name=fixture_name)
    fixture2.append_line('assert this == slash.context.fixture')
    summary = suite.run()
    assert len(summary.session.warnings) == 1
    assert 'Duplicate' in summary.session.warnings.warnings[0].details['message']
    assert fixture_name in summary.session.warnings.warnings[0].details['message']


@pytest.mark.parametrize('specific_method', [True, False])
@pytest.mark.parametrize('with_parameters', [True, False])
def test_iter_specific_factory(suite, suite_test, specific_method, with_parameters):

    if suite_test.cls is not None and specific_method:
        suite_test.cls.add_method_test()

    if with_parameters:
        suite_test.add_parameter()

    for test in suite:
        if suite_test.cls is None and test is not suite_test:
            # we are selecting a specific function, and that's not it:
            test.expect_deselect()
        elif suite_test.cls is not None and test.cls is not suite_test.cls:
            test.expect_deselect()
        elif specific_method and suite_test.cls is test.cls and suite_test is not test:
            test.expect_deselect()

    path = suite.commit()
    if suite_test.cls:
        assert suite_test.cls.tests
        factory_name = suite_test.cls.name
    else:
        factory_name = suite_test.name

    pattern = '{}:{}'.format(os.path.join(path, suite_test.file.get_relative_path()), factory_name)
    if suite_test.cls is not None and specific_method:
        assert len(suite_test.cls.tests) > 1
        pattern += '.{}'.format(suite_test.name)
    suite.run(args=[pattern])


def test_import_error_registers_as_session_error(active_slash_session, test_loader):
    with pytest.raises(CannotLoadTests):
        test_loader.get_runnables(["/non/existent/path"])
    errors = active_slash_session.results.global_result.get_errors()
    assert len(errors) == 1
    [error] = errors  # pylint: disable=unused-variable

def test_no_traceback_for_slash_exception():
    suite = Suite()
    summary = suite.run(expect_session_errors=True)
    assert not summary.session.results.is_success()
    [err] = summary.session.results.global_result.get_errors()
    assert err.exception_type is CannotLoadTests
    output = summary.get_console_output()
    assert 'Traceback' not in output

def test_no_traceback_for_marked_exceptions():
    suite = Suite()

    @suite.slashconf.append_body
    def __code__():  # pylint: disable=unused-variable
        from slash.exception_handling import inhibit_unhandled_exception_traceback
        raise inhibit_unhandled_exception_traceback(Exception('Some Error'))

    summary = suite.run(expect_session_errors=True)
    assert not summary.session.results.is_success()
    errors = summary.session.results.global_result.get_errors()
    assert [err.exception_type for err in errors] == [Exception]
    assert 'Some Error' in errors[0].exception_str
    output = summary.get_console_output()
    assert 'Traceback' not in output

def test_import_errors_with_session():

    suite = Suite()

    for _ in range(20):
        suite.add_test()

    problematic = suite.files[1]
    problematic.prepend_line('from nonexistent import nonexistent')

    for test in suite:
        test.expect_deselect()

    summary = suite.run(expect_session_errors=True)

    assert summary.exit_code != 0

    errs = summary.session.results.global_result.get_errors()
    for err in errs:
        assert 'No module named nonexistent' in err.message or "No module named 'nonexistent'" in err.message


    return suite



import sys
import errno
import os
import signal
import subprocess
import time
import logbook
import threading
from tempfile import mkdtemp
from  six.moves import xmlrpc_client
from .. import log
from ..exceptions import INTERRUPTION_EXCEPTIONS, ParallelServerIsDown, ParallelTimeout
from ..conf import config
from ..ctx import context
from .server import Server, ServerStates
from ..utils.tmux_utils import create_new_window, create_new_pane
from .worker_configuration import WorkerConfiguration
from .._compat import iteritems
from .. import hooks

_logger = logbook.Logger(__name__)
log.set_log_color(_logger.name, logbook.NOTICE, 'blue')

TIME_BETWEEN_CHECKS = 2
MAX_CONNECTION_RETRIES = 200

class ParallelManager(object):
    def __init__(self, args):
        super(ParallelManager, self).__init__()
        self.server = None
        self.workers_error_dircetory = mkdtemp()
        self.args = [sys.executable, '-m', 'slash.frontend.main', 'run', '--parallel-parent-session-id', context.session.id, \
                    '--workers-error-dir', self.workers_error_dircetory] + args
        self.workers_num = config.root.parallel.num_workers
        self.workers = {}
        self.max_worker_id = 1
        self.server_thread = None

    def try_connect(self):
        num_retries = 0
        while self.server.state == ServerStates.NOT_INITIALIZED:
            time.sleep(0.1)
            if num_retries == MAX_CONNECTION_RETRIES:
                raise ParallelServerIsDown("Cannot connect to XML_RPC server")
            num_retries += 1

    def start_worker(self):
        worker_id = str(self.max_worker_id)
        _logger.notice("Starting worker number {}", worker_id)
        new_args = self.args[:] + ["--parallel-worker-id", worker_id]
        worker_config = WorkerConfiguration(new_args)
        hooks.before_worker_start(worker_config=worker_config) # pylint: disable=no-member
        if config.root.tmux.enabled:
            worker_config.argv.append(';$SHELL')
            command = ' '.join(worker_config.argv)
            if config.root.tmux.use_panes:
                self.workers[worker_id] = create_new_pane(command)
            else:
                self.workers[worker_id] = create_new_window("worker {}".format(worker_id), command)
        else:
            with open(os.devnull, 'w') as devnull:
                proc = subprocess.Popen(worker_config.argv, stdin=devnull, stdout=devnull, stderr=devnull)
                self.workers[worker_id] = proc
        _logger.trace("Worker #{} command: {}", worker_id, ' '.join(worker_config.argv))
        self.max_worker_id += 1


    def start_server_in_thread(self, collected):
        self.server = Server(collected)
        self.server_thread = threading.Thread(target=self.server.serve, args=())
        self.server_thread.setDaemon(True)
        self.server_thread.start()

    def get_proxy(self):
        return xmlrpc_client.ServerProxy('http://{}:{}'.format(config.root.parallel.server_addr, self.server.port))

    def kill_workers(self):
        if config.root.tmux.enabled:
            for worker_pid in self.server.worker_pids:
                try:
                    os.kill(worker_pid, signal.SIGINT)
                except OSError as err:
                    if err.errno != errno.ESRCH:
                        raise
        else:
            for worker in self.workers.values():
                worker.send_signal(signal.SIGINT)

    def report_worker_error_logs(self):
        found_worker_errors_file = False
        for file_name in os.listdir(self.workers_error_dircetory):
            if file_name.startswith(config.root.parallel.worker_error_file):
                found_worker_errors_file = True
                with open(os.path.join(self.workers_error_dircetory, file_name)) as worker_file:
                    content = worker_file.readlines()
                    for line in content:
                        _logger.error("{}: {}", file_name, line)
        if not found_worker_errors_file:
            _logger.error("No worker error files were found")


    def wait_all_workers_to_connect(self):
        while self.server.state == ServerStates.WAIT_FOR_CLIENTS:
            if time.time() - self.server.start_time > config.root.parallel.worker_connect_timeout * self.workers_num:
                _logger.error("Timeout: Not all clients connected to server, terminating")
                _logger.error("Clients connected: {}", self.server.clients_last_communication_time.keys())
                self.kill_workers()
                self.report_worker_error_logs()
                raise ParallelTimeout("Not all clients connected")
            time.sleep(TIME_BETWEEN_CHECKS)

    def check_worker_timed_out(self):
        for worker_id, last_connection_time in iteritems(self.server.get_workers_last_connection_time()):
            if time.time() - last_connection_time > config.root.parallel.communication_timeout_secs:
                _logger.error("Worker {} is down, terminating session", worker_id)
                self.report_worker_error_logs()
                if not config.root.tmux.enabled:
                    if self.workers[worker_id].poll() is None:
                        self.workers[worker_id].kill()
                elif not config.root.tmux.use_panes:
                    self.workers[worker_id].rename_window('stopped_client_{}'.format(worker_id))
                self.get_proxy().report_client_failure(worker_id)

    def check_no_requests_timeout(self):
        if time.time() - self.server.last_request_time > config.root.parallel.no_request_timeout:
            _logger.error("No request sent to server for {} seconds, terminating",
                          config.root.parallel.no_request_timeout)
            if self.server.has_connected_clients():
                _logger.error("Clients that are still connected to server: {}",
                              self.server.clients_last_communication_time.keys())
            if self.server.has_more_tests():
                _logger.error("Number of unstarted tests: {}", self.server.unstarted_tests.qsize())
            if self.server.executing_tests:
                _logger.error("Currently executed tests indexes: {}", self.server.executing_tests.values())
            self.kill_workers()
            self.report_worker_error_logs()
            raise ParallelTimeout("No request sent to server for {} seconds".format(config.root.parallel.no_request_timeout))

    def is_process_running(self, pid):
        try:
            os.kill(pid, 0)
        except OSError as err:
            if err.errno == errno.ESRCH:
                return False
            else:
                raise
        return True

    def start(self):
        self.try_connect()
        if not config.root.parallel.server_port:
            self.args.extend(['--parallel-port', str(self.server.port)])
        try:
            for _ in range(self.workers_num):
                self.start_worker()
            self.wait_all_workers_to_connect()
            while self.server.should_wait_for_request():
                self.check_worker_timed_out()
                self.check_no_requests_timeout()
                time.sleep(TIME_BETWEEN_CHECKS)
        except INTERRUPTION_EXCEPTIONS:
            _logger.error("Server interrupted, stopping workers and terminating")
            self.get_proxy().session_interrupted()
            self.kill_workers()
            raise
        finally:
            if not config.root.tmux.enabled:
                for worker in self.workers.values():
                    worker.wait()
            else:
                for worker_pid in self.server.worker_pids:
                    for _ in range(10):
                        if not self.is_process_running(worker_pid):
                            break
                        else:
                            time.sleep(0.5)
            self.get_proxy().stop_serve()
            self.server_thread.join()



# pylint: disable=len-as-condition,redefined-outer-name
import pytest

from .utils.suite_writer import Suite


def test_autouse_fixtures_global(tracked_suite):

    fixture = tracked_suite.slashconf.add_fixture(autouse=True)

    results = tracked_suite.run().session.results
    assert len(results) > 0
    for result in results:
        assert fixture.id in result.data['active_fixtures']

def test_autouse_fixtures_specific_module(tracked_suite, suite_test):

    fixture = suite_test.file.add_fixture(autouse=True)
    summary = tracked_suite.run()
    assert len(summary.session.results) > 0
    assert len(tracked_suite.files) > 1

    for test in tracked_suite:
        for result in summary.get_all_results_for_test(test):
            if test.file is suite_test.file:
                assert fixture.id in result.data['active_fixtures']
            else:
                assert fixture.id not in result.data['active_fixtures']

@pytest.mark.parametrize('scope', ['test', 'session', 'module'])
@pytest.mark.parametrize('depend_explicitly', [True, False])
def test_autouse_called_first(scope, test_type, depend_explicitly):
    suite = Suite()
    suite_test = suite.add_test(type=test_type)

    autouse_fixture = suite.slashconf.add_fixture(autouse=True, scope=scope)
    autouse_fixture_called = autouse_fixture.add_event()

    regular_fixture = suite.slashconf.add_fixture()
    regular_fixture_called = regular_fixture.add_event()
    suite_test.depend_on_fixture(regular_fixture)

    if depend_explicitly:
        suite_test.depend_on_fixture(autouse_fixture)

    test_called = suite_test.add_event()
    events = suite.run().events
    assert events[regular_fixture_called].timestamp < events[test_called].timestamp
    assert events[autouse_fixture_called].timestamp < events[regular_fixture_called].timestamp


@pytest.fixture  # pylint: disable=unused-argument
def tracked_suite(suite, suite_test):  # pylint: disable=unused-argument
    assert len(suite)
    for test in suite:
        test.prepend_line('slash.context.result.data["active_fixtures"] = __ut__.get_fixture_memento()')
    return suite



from .test import Test


class NonMethodTest(Test):  # pylint: disable=abstract-method

    cls = None

    def __init__(self, suite, file):  # pylint: disable=useless-super-delegation
        super(NonMethodTest, self).__init__(suite, file)

    def is_method_test(self):
        return False



from uuid import uuid4
import itertools

from .function import Function


class Fixture(Function):

    def __init__(self, suite, file, scope=None, autouse=False, name=None):
        if name is None:
            name = 'fx_{}'.format(str(uuid4()).replace('-', '')[:6])
        super(Fixture, self).__init__(suite, name=name)
        self.file = file
        self.scope = scope
        self.autouse = autouse

    def is_generator_fixture(self):
        return False

    def add_cleanup(self, **kwargs):
        return self.add_deferred_event('this.add_cleanup', name='fixture_cleanup', **kwargs)

    def _write_decorators(self, code_formatter):
        self._write_fixture_decorator(code_formatter)
        super(Fixture, self)._write_decorators(code_formatter)

    def _write_prologue(self, code_formatter):
        if not self.suite.debug_info:
            return

        self._write_event(code_formatter, 'fixture_start')
        code_formatter.writeln(
            '__ut__.notify_fixture_start({!r}, {})'.format(self.id, self.get_value_string()))
        code_formatter.writeln('@this.add_cleanup')
        code_formatter.writeln('def cleanup():')
        with code_formatter.indented():
            self._write_event(code_formatter, 'fixture_end')
            code_formatter.writeln(
                '__ut__.notify_fixture_end({!r})'.format(self.id))

        super(Fixture, self)._write_prologue(code_formatter)

    def _write_return(self, code_formatter):
        code_formatter.writeln(
            'return {}'.format(self.get_value_string()))

    def _write_fixture_decorator(self, code_formatter):

        params = {}
        if self.scope is not None:
            params['scope'] = self.scope
        if self.autouse:
            params['autouse'] = True

        code_formatter.write('@slash.fixture')
        if params:
            code_formatter.write('({})'.format(', '.join(
                '{}={!r}'.format(k, v) for k, v in params.items())))
        code_formatter.writeln()

    def get_value_string(self):
        returned = '{{"value": {!r}, "params": {{'.format(self.name)

        for name, param in itertools.chain(
                ((p.name, p) for p in self._parameters),
                ((alias or f.name, f) for alias, f, _ in self._fixtures),
        ):
            returned += '{!r}: {},'.format(param.id, name)
        returned += '} }'
        return returned

    def _get_argument_strings(self):
        return itertools.chain(['this'], super(Fixture, self)._get_argument_strings())

    def __repr__(self):
        return '<Fixture {}>'.format(self.name)



# pylint: disable=redefined-outer-name
import pytest
import slash.resuming
from slash.resuming import (CannotResume, get_last_resumeable_session_id, get_tests_from_previous_session)


def test_resume_no_session():
    with pytest.raises(CannotResume):
        get_tests_from_previous_session("nonexisting_session")

def test_get_last_resumeable_session(suite):
    suite[len(suite) // 2].when_run.fail()
    prev_id = None
    for i in range(5):  # pylint: disable=unused-variable
        results = suite.run()
        assert results.session.id != prev_id
        prev_id = results.session.id
        assert get_last_resumeable_session_id() == results.session.id

def test_resume(suite):
    fail_index = len(suite) // 2
    suite[fail_index].when_run.fail()
    for index, test in enumerate(suite):
        if index > fail_index:
            test.expect_not_run()
    result = suite.run(additional_args=['-x'])
    resumed = get_tests_from_previous_session(result.session.id)

    assert len(resumed) + result.session.results.get_num_started() - 1 == len(suite)

def test_resume_with_parametrization(suite, suite_test):
    num_values1 = 3
    num_values2 = 5
    suite_test.add_parameter(num_values=num_values1)
    suite_test.add_parameter(num_values=num_values2)
    fail_index = len(suite) // 2
    suite[fail_index].when_run.fail()
    summary = suite.run()
    resumed = get_tests_from_previous_session(summary.session.id)

    assert len(summary.get_all_results_for_test(suite_test)) == num_values1 * num_values2
    assert len(resumed) == 1
    assert resumed[0].function_name == suite[fail_index].name

def test_different_folder_no_resume_session_id(suite, tmpdir):  # pylint: disable=unused-argument
    fail_index = len(suite) // 2
    suite[fail_index].when_run.fail()
    suite.run()
    sessoin_id = get_last_resumeable_session_id()

    assert sessoin_id
    with tmpdir.ensure_dir().as_cwd():
        with pytest.raises(CannotResume):
            sessoin_id = get_last_resumeable_session_id()

def test_delete_old_sessions(suite):
    result = suite.run()
    assert result.session.id == get_last_resumeable_session_id()
    slash.resuming._MAX_DAYS_SAVED_SESSIONS = 0 # pylint: disable=protected-access
    result = suite.run()
    with pytest.raises(CannotResume):
        get_last_resumeable_session_id()



import copy
import logbook
import functools
import time
from enum import Enum
from six.moves import queue
from six.moves import xmlrpc_server
from ..utils.python import unpickle
from .. import log
from ..ctx import context
from ..runner import _get_test_context
from .. import hooks
from ..conf import config

_logger = logbook.Logger(__name__)
log.set_log_color(_logger.name, logbook.NOTICE, 'blue')

NO_MORE_TESTS = "NO_MORE_TESTS"
PROTOCOL_ERROR = "PROTOCOL_ERROR"
WAITING_FOR_CLIENTS = "WAITING_FOR_CLIENTS"

class ServerStates(Enum):
    NOT_INITIALIZED = 1
    WAIT_FOR_CLIENTS = 2
    WAIT_FOR_COLLECTION_VALIDATION = 3
    SERVE_TESTS = 4
    STOP_TESTS_SERVING = 5
    STOP_SERVE = 6

def server_func(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        self = args[0]
        client_id = kwargs.pop('client_id', args[1])
        self.clients_last_communication_time[client_id] = self.last_request_time = time.time()
        return func(*args)  # pylint: disable=not-callable
    return wrapper

class Server(object):
    def __init__(self, tests):
        super(Server, self).__init__()
        self.host = config.root.parallel.server_addr
        self.worker_session_error_reported = False
        self.interrupted = False
        self.state = ServerStates.NOT_INITIALIZED
        self.port = None
        self.tests = tests
        self.worker_session_ids = []
        self.executing_tests = {}
        self.finished_tests = []
        self.unstarted_tests = queue.Queue()
        self.last_request_time = time.time()
        self.num_collections_validated = 0
        self.start_time = time.time()
        self.worker_pids = []
        for i in range(len(tests)):
            self.unstarted_tests.put(i)
        self.clients_last_communication_time = {}
        self.collection = [[test.__slash__.file_path, test.__slash__.function_name, test.__slash__.variation.id] for test in self.tests]

    def get_workers_last_connection_time(self):
        return copy.deepcopy(self.clients_last_communication_time)

    def _has_unstarted_tests(self):
        return not self.unstarted_tests.empty()

    def has_connected_clients(self):
        return len(self.clients_last_communication_time) > 0

    def has_more_tests(self):
        return len(self.finished_tests) < len(self.tests)

    def report_client_failure(self, client_id):
        self.clients_last_communication_time.pop(client_id)
        test_index = self.executing_tests.get(client_id, None)
        if test_index is not None:
            _logger.error("Worker {} interrupted while executing test {}", client_id,
                          self.tests[test_index].__slash__.address)
            with _get_test_context(self.tests[test_index], logging=False) as (result, _):
                result.mark_interrupted()
                self.finished_tests.append(test_index)
        self.state = ServerStates.STOP_TESTS_SERVING
        self._mark_unrun_tests()

    def _mark_unrun_tests(self):
        while self._has_unstarted_tests():
            test_index = self.unstarted_tests.get()
            with _get_test_context(self.tests[test_index], logging=False):
                pass
            self.finished_tests.append(test_index)

    def _get_worker_session_id(self, client_id):
        return "worker_{}".format(client_id)

    @server_func
    def keep_alive(self, client_id):
        _logger.debug("Client_id {} sent keep_alive", client_id)

    @server_func
    def connect(self, client_id, client_pid):
        _logger.notice("Client_id {} connected", client_id)
        client_session_id = '{}_{}'.format(context.session.id.split('_')[0], client_id)
        context.session.logging.create_worker_symlink(self._get_worker_session_id(client_id), client_session_id)
        hooks.worker_connected(session_id=client_session_id)  # pylint: disable=no-member
        self.worker_session_ids.append(client_session_id)
        self.worker_pids.append(client_pid)
        self.executing_tests[client_id] = None
        if len(self.clients_last_communication_time) >= config.root.parallel.num_workers:
            _logger.notice("All workers connected to server")
            self.state = ServerStates.WAIT_FOR_COLLECTION_VALIDATION

    @server_func
    def validate_collection(self, client_id, client_collection):
        if not self.collection == client_collection:
            _logger.error("Client_id {} sent wrong collection", client_id)
            return False
        self.num_collections_validated += 1
        _logger.debug("Worker {} validated tests successfully", client_id)
        if self.num_collections_validated >= config.root.parallel.num_workers and self.state == ServerStates.WAIT_FOR_COLLECTION_VALIDATION:
            _logger.notice("All workers collected tests successfully, start serving tests")
            self.state = ServerStates.SERVE_TESTS
        return True

    @server_func
    def disconnect(self, client_id):
        _logger.notice("Client {} sent disconnect", client_id)
        self.clients_last_communication_time.pop(client_id)
        self.state = ServerStates.STOP_TESTS_SERVING

    @server_func
    def get_test(self, client_id):
        if not self.executing_tests[client_id] is None:
            _logger.error("Client_id {} requested new test without sending former result", client_id)
            return PROTOCOL_ERROR
        if self.state == ServerStates.STOP_TESTS_SERVING:
            return NO_MORE_TESTS
        elif self.state in [ServerStates.WAIT_FOR_CLIENTS, ServerStates.WAIT_FOR_COLLECTION_VALIDATION]:
            return WAITING_FOR_CLIENTS
        elif self.state == ServerStates.SERVE_TESTS and self._has_unstarted_tests():
            test_index = self.unstarted_tests.get()
            test = self.tests[test_index]
            self.executing_tests[client_id] = test_index
            hooks.test_distributed(test_logical_id=test.__slash__.id, worker_session_id=self._get_worker_session_id(client_id)) # pylint: disable=no-member
            _logger.notice("#{}: {}, Client_id: {}", test_index + 1, test.__slash__.address, client_id,
                           extra={'highlight': True, 'filter_bypass': True})
            return test_index
        else:
            _logger.debug("No unstarted tests, sending end to client_id {}", client_id)
            self.state = ServerStates.STOP_TESTS_SERVING
            return NO_MORE_TESTS

    @server_func
    def finished_test(self, client_id, result_dict):
        _logger.debug("Client_id {} finished_test", client_id)
        test_index = self.executing_tests.get(client_id, None)
        if test_index is not None:
            self.finished_tests.append(test_index)
            self.executing_tests[client_id] = None
            with _get_test_context(self.tests[test_index], logging=False) as (result, _):
                result.deserialize(result_dict)
                context.session.reporter.report_test_end(self.tests[test_index], result)
                if not result.is_success(allow_skips=True) and config.root.run.stop_on_error:
                    _logger.debug("Stopping (run.stop_on_error==True)")
                    self.state = ServerStates.STOP_TESTS_SERVING
                    self._mark_unrun_tests()
        else:
            _logger.error("finished_test request from client_id {} with index {}, but no test is mapped to this worker",
                          client_id, test_index)
            return PROTOCOL_ERROR

    def stop_serve(self):
        self.state = ServerStates.STOP_SERVE

    def session_interrupted(self):
        self.interrupted = True
        if self.state != ServerStates.STOP_SERVE:
            self.state = ServerStates.STOP_TESTS_SERVING

    @server_func
    def report_warning(self, client_id, pickled_warning):
        _logger.notice("Client_id {} sent warning", client_id)
        try:
            warning = unpickle(pickled_warning)
            context.session.warnings.add(warning)
        except TypeError:
            _logger.error('Error when deserializing warning, not adding it')

    @server_func
    def report_session_error(self, client_id, session_error):
        self.worker_session_error_reported = True
        _logger.error("Client_id {} sent session error: {}", client_id, session_error)

    def should_wait_for_request(self):
        return self.has_connected_clients() or self.has_more_tests()

    def serve(self):
        server = xmlrpc_server.SimpleXMLRPCServer((self.host, config.root.parallel.server_port), allow_none=True, logRequests=False)
        try:
            self.port = server.server_address[1]
            self.state = ServerStates.WAIT_FOR_CLIENTS
            server.register_instance(self)
            _logger.debug("Starting server loop")
            while self.state != ServerStates.STOP_SERVE:
                server.handle_request()
            if not self.interrupted:
                context.session.mark_complete()
            _logger.trace('Session finished. is_success={0} has_skips={1}',
                          context.session.results.is_success(allow_skips=True), bool(context.session.results.get_num_skipped()))
            _logger.debug("Exiting server loop")
        finally:
            server.server_close()



class FixtureBase(object):

    info = None
    keyword_arguments = None
    parametrization_ids = None
    names = None

    def get_value(self, kwargs, active_fixture):
        raise NotImplementedError()  # pragma: no cover

    def get_variations(self):
        return None  # pragma: no cover

    def resolve(self, store):
        if self.keyword_arguments is None:
            self.keyword_arguments = self._resolve(store)

    def _resolve(self, store):
        raise NotImplementedError()  # pragma: no cover



import functools
import itertools
import inspect

from sentinels import NOTHING
import six

from ...ctx import context
from ..._compat import izip, iteritems
from ...utils.python import get_arguments_dict, wraps
from ...utils.function_marker import function_marker

_id_gen = itertools.count(1000)


def _ensure_fixture_info(func, **kwargs):
    if not hasattr(func, '__slash_fixture__'):
        func.__slash_fixture__ = FixtureInfo(func, **kwargs)
    return func

def fixture(func=None, name=None, scope=None, autouse=False):
    if func is None:
        return functools.partial(fixture, name=name, scope=scope, autouse=autouse)

    if inspect.isgeneratorfunction(func):
        return yield_fixture(func=func, name=name, scope=scope, autouse=autouse)
    return _ensure_fixture_info(func=func, name=name, scope=scope, autouse=autouse)

nofixtures = function_marker('__slash_nofixtures__')
nofixtures.__doc__ = 'Marks the decorated function as opting out of automatic fixture deduction. ' + \
                     'Slash will not attempt to parse needed fixtures from its argument list'


class FixtureInfo(object):

    def __init__(self, func=None, name=None, scope=None, autouse=False, path=None):
        super(FixtureInfo, self).__init__()
        self.path = path
        self.id = next(_id_gen)
        if name is None:
            if func is None:
                name = '__unnamed_{}'.format(self.id)
            else:
                name = func.__name__
        if scope is None:
            scope = 'test'
        self.name = name
        self.func = func
        self.autouse = autouse
        self.scope = _SCOPES[scope]
        if self.func is not None:

            self.required_args = get_arguments_dict(self.func)
        else:
            self.required_args = {}
        if 'this' in self.required_args:
            self.required_args.pop('this')
            self.needs_this = True
        else:
            self.needs_this = False

    @property
    def scope_name(self):
        return _SCOPES_BY_ID[self.scope]

def get_scope_by_name(scope_name):
    return _SCOPES[scope_name]

def get_scope_name_by_scope(scope_id):
    return _SCOPES_BY_ID[scope_id]

_SCOPES = dict(
    izip(('test', 'module', 'session'), itertools.count()))

_SCOPES_BY_ID = dict((id, name) for (name, id) in iteritems(_SCOPES))


def generator_fixture(func=None, **kw):
    """A utility for generating parametrization values from a generator:

    >>> @slash.generator_fixture
    ... def some_parameter():
    ...     yield first_value
    ...     yield second_value

    .. note:: A generator parameter is a shortcut for a simple parametrized fixture, so the entire iteration is exhausted during test load time
    """
    from .parameters import parametrize

    if func is None:
        return functools.partial(generator_fixture, **kw)

    @parametrize('param', list(func()))
    def new_func(param):
        return param

    kw.setdefault('name', func.__name__)
    return _ensure_fixture_info(func=new_func, **kw)


def yield_fixture(func=None, **kw):
    """Builds a fixture out of a generator. The pre-yield part of the generator is used as the setup, where the
    yielded value becomes the fixture value. The post-yield part is added as a cleanup:

    >>> @slash.yield_fixture
    ... def some_fixture(arg1, arg2):
    ...     m = Microwave()
    ...     m.turn_on(wait=True)
    ...     yield m
    ...     m.turn_off()
    """

    if func is None:
        return functools.partial(yield_fixture, **kw)

    func = _ensure_fixture_info(func=func, **kw)

    @wraps(func)
    def new_func(**kwargs):
        f = func(**kwargs)
        value = next(f)
        @context.fixture.add_cleanup
        def cleanup(): # pylint: disable=unused-variable
            try:
                next(f)
            except StopIteration:
                pass
            else:
                raise RuntimeError('Yielded fixture did not stop at cleanup')
        return value
    return new_func



class _use_type(type):

    def __getattr__(cls, attr):
        if attr.startswith('_'):
            raise AttributeError(attr)
        return cls(attr)


class use(six.with_metaclass(_use_type)): # pylint: disable=no-init
    """
    Allows tests to use fixtures under different names:

    >>> def test_something(m: use('microwave')):
    ...    ...

    For cosmetic purposes, you can also use ``use`` with attribute access:

    >>> def test_something(m: use.microwave):
    ...    ...

    """
    def __init__(self, real_fixture_name):
        self.real_fixture_name = real_fixture_name

    def __repr__(self):
        return '<slash.use({!r})>'.format(self.real_fixture_name)


def get_real_fixture_name_from_argument(argument):
    if argument.annotation is not NOTHING and isinstance(argument.annotation, use):
        return argument.annotation.real_fixture_name
    return argument.name


__all__ = ['fixture', 'nofixtures', 'generator_fixture', 'yield_fixture', 'use']



import sys


class warning_callback_context(object):

    """A context for installing callbacks for handling warnings.

    Code adopted from the built-in catch_warnings handler
    """

    def __init__(self, callback):
        self._callback = callback
        self._module = sys.modules['warnings']
        self._entered = False

    def __enter__(self):
        if self._entered:
            raise RuntimeError("Cannot enter %r twice" % self)
        self._entered = True
        self._showwarning = self._module.showwarning
        self._module.showwarning = self._callback

    def __exit__(self, *exc_info):
        if not self._entered:
            raise RuntimeError("Cannot exit %r without entering first" % self)
        self._module.showwarning = self._showwarning



#!/usr/bin/env python
import argparse
import sys

import random

from tests.utils.suite_writer import Suite

parser = argparse.ArgumentParser(usage="%(prog)s [options] args...")
parser.add_argument('--fixtures', dest='use_fixtures', action='store_true', default=False)
parser.add_argument('--parameters', dest='use_parameters', action='store_true', default=False)
parser.add_argument('--debug-info', dest='debug_info', action='store_true', default=False)
parser.add_argument("dir")
parser.add_argument("summary")

_FIXTURE_FREQ = 3
_PARAM_FREQ = 4

class Application(object):

    _INDENT = " " * 4

    def __init__(self, args):
        self._args = args

    def main(self):

        s = Suite(path=self._args.dir, debug_info=self._args.debug_info)
        if not self._args.summary:
            parser.error("No summary given")
        for index, element in enumerate(self._args.summary):
            t = s.add_test()
            if self._args.use_fixtures and index % _FIXTURE_FREQ == 0:
                if index % 2 == 0:
                    t.depend_on_fixture(t.file.add_fixture())
                else:
                    t.depend_on_fixture(s.slashconf.add_fixture())

            if self._args.use_parameters and index % _PARAM_FREQ == 0:
                t.parametrize()

            if element == '.':
                pass
            elif element == 'F':
                t.when_run.fail()
            elif element == 'E':
                t.when_run.error()
            elif element == 'S':
                t.when_run.skip()
            elif element == 'i':
                t.when_run.interrupt()
            else:
                parser.error("Unknown marker: {!r}".format(element))

        s.commit()

        return 0


#### For use with entry_points/console_scripts
def main_entry_point():
    args = parser.parse_args()
    app = Application(args)
    sys.exit(app.main())


if __name__ == "__main__":
    main_entry_point()



from uuid import uuid4
import pytest
import logbook

@pytest.mark.parametrize('failure_type', ['failure', 'error'])
@pytest.mark.parametrize('use_custom_message', [True, False])
def test_add_error_custom_exc_info(suite, suite_test, failure_type, use_custom_message):
    message = str(uuid4())

    if use_custom_message:
        add_error_args = '{!r}, '.format(message)
    else:
        add_error_args = ''

    add_error_args += 'exc_info=exc_info'

    code = """
import sys
try:
    1/0
except ZeroDivisionError:
    exc_info = sys.exc_info()
    try:
        None.a = 2
    except AttributeError:
        slash.add_{0}({1})
    """.format(failure_type, add_error_args)

    for line in code.strip().splitlines():
        suite_test.append_line(line)

    if failure_type == 'error':
        suite_test.expect_error()
    else:
        suite_test.expect_failure()
    summary = suite.run()
    [result] = summary.get_all_results_for_test(suite_test)

    if failure_type == 'error':
        [err] = result.get_errors()
    else:
        [err] = result.get_failures()

    assert err.has_custom_message() == use_custom_message
    if use_custom_message:
        assert err.message == message
    assert err.exception_type is ZeroDivisionError
    assert err.traceback.frames


def test_add_fatal_error(suite, suite_test):

    @suite_test.append_body
    def __code__():             # pylint: disable=unused-variable
        slash.add_error('bla').mark_fatal() # pylint: disable=undefined-variable

    suite_test.expect_error()
    for test in suite.iter_all_after(suite_test):
        test.expect_not_run()

    session = suite.run().session
    assert session.results.has_fatal_errors()


def test_session_level_add_error_message(suite, suite_test):
    @suite_test.file.append_body
    def __code__():                       # pylint: disable=unused-variable
        @slash.hooks.session_end.register # pylint: disable=undefined-variable
        def _callback():
            slash.add_error('session: add_error') # pylint: disable=undefined-variable
    res = suite.run(expect_session_errors=True)
    errors = res.session.results.global_result.get_errors()
    assert len(errors) == 1
    [err] = errors
    assert err.message == 'session: add_error'


@pytest.mark.parametrize('log_variables', [True, False])
def test_add_error_log_traceback_variables(suite, suite_test, log_variables, config_override, tmpdir):
    config_override('log.core_log_level', logbook.TRACE)
    config_override('log.traceback_variables', log_variables)
    config_override('log.root', str(tmpdir.join('logs')))

    @suite_test.prepend_body
    def __code__():          # pylint: disable=unused-variable
        # to avoid the line itself from being detected
        x_variable = 'x' * 3 # pylint: disable=unused-variable
        class Object(object):

            def __init__(self):
                self.property_value = 'yyy'

        self = Object() # pylint: disable=unused-variable

    suite_test.when_run.error()
    res = suite.run()
    result = res[suite_test]

    with open(result.get_log_path()) as f:
        lines = f.readlines()

    def _search_variable(variable_name, variable_value):
        found = False
        for line in lines:
            if variable_name in line and variable_value in line:
                found = True
                break
        assert found == log_variables, 'Variable {!r} not found in traceback log!'.format(variable_name)
    _search_variable('x_variable', 'xxx')
    _search_variable('self.property_value', 'yyy')


def test_add_error_log_traceback_variables_self_none(suite, suite_test, config_override, tmpdir):
    config_override('log.core_log_level', logbook.TRACE)
    config_override('log.traceback_variables', True)
    config_override('log.root', str(tmpdir.join('logs')))

    @suite_test.prepend_body
    def __code__():          # pylint: disable=unused-variable
        # to avoid the line itself from being detected
        self = None # pylint: disable=unused-variable

    suite_test.when_run.error()
    res = suite.run()
    result = res[suite_test]

    with open(result.get_log_path()) as f:
        lines = f.read()

    assert 'self: None' in lines
    [err] = result.get_errors()
    assert 'Test exception' in str(err)



"""
This module contains the implementation of a traceback.traceback object proxy.
It is needed in order to make creation and modification of these objects possible,
which is not supported in Python by default, as they are quite internal types.
It lets us concatenate and fake tracebacks in case of handling an exception.
The manipulation of tracebacks is done using ctypes hacks and casting pointers to structs,
and works only with CPython.
"""

import sys
import traceback
import types
import ctypes
import inspect

from .._compat import PY2, PYPY

__all__ = ["create_traceback_proxy"]

if PYPY:
    def create_traceback_proxy(tb=None, frame_correction=0): # pylint: disable=unused-argument
        raise NotImplementedError("Tracebacks manipulation is not possible in PyPy")

else:
    class TracebackProxy(object):
        """
        Wraps the builtin traceback.traceback object.
        Exports exactly the same interface, making these types interchangable
        """

        if PY2:
            if hasattr(ctypes.pythonapi, 'Py_InitModule4_64'):
                _Py_ssize_t = ctypes.c_int64
            else:
                _Py_ssize_t = ctypes.c_int
        else:
            _Py_ssize_t = ctypes.c_ssize_t

        class _PyObject(ctypes.Structure):
            pass

        # Python build with "--with-pydebug"
        if hasattr(sys, 'getobjects'):
            _PyObject._fields_ = [ # pylint: disable=protected-access
                ('_ob_next', ctypes.POINTER(_PyObject)),
                ('_ob_prev', ctypes.POINTER(_PyObject)),
                ('ob_refcnt', _Py_ssize_t),
                ('ob_type', ctypes.POINTER(_PyObject))
            ]
        else:
            _PyObject._fields_ = [ # pylint: disable=protected-access
                ('ob_refcnt', _Py_ssize_t),
                ('ob_type', ctypes.POINTER(_PyObject))
            ]

        class _Frame(_PyObject):
            """
            Represents a traceback.frame object
            """
            pass

        class _Traceback(_PyObject):
            """
            Represents a traceback.traceback object
            """
            pass

        _Traceback._fields_ = [ # pylint: disable=protected-access
            ('tb_next', ctypes.POINTER(_Traceback)),
            ('tb_frame', ctypes.POINTER(_Frame)),
            ('tb_lasti', ctypes.c_int),
            ('tb_lineno', ctypes.c_int)
        ]

        def __init__(self, tb=None, frame=None):
            assert tb is not None or frame is not None
            self._tb = TracebackProxy.create_traceback()
            self._obj = TracebackProxy._Traceback.from_address(id(self._tb)) # pylint: disable=no-member
            self.tb_next = None
            if tb:
                self.tb_frame = tb.tb_frame
                self.tb_lasti = tb.tb_lasti
                self.tb_lineno = tb.tb_lineno
            else:
                self.tb_frame = frame
                self.tb_lasti = frame.f_lasti
                self.tb_lineno = frame.f_lineno

        def print_tb(self):
            traceback.print_tb(self._tb)

        @property
        def tb_next(self):
            return self._tb.tb_next

        @tb_next.setter
        def tb_next(self, tb):
            if self._tb.tb_next:
                old = TracebackProxy._Traceback.from_address(id(self._tb.tb_next)) # pylint: disable=no-member
                old.ob_refcnt -= 1

            assert tb is None or isinstance(tb, types.TracebackType) or isinstance(tb, TracebackProxy)
            if tb:
                obj = TracebackProxy._Traceback.from_address(id(tb)) # pylint: disable=no-member
                obj.ob_refcnt += 1
                self._obj.tb_next = ctypes.pointer(obj)
            else:
                self._obj.tb_next = ctypes.POINTER(TracebackProxy._Traceback)()

        @property
        def tb_frame(self):
            return self._tb.tb_frame

        @tb_frame.setter
        def tb_frame(self, frame):
            if self._tb.tb_frame:
                old = TracebackProxy._Frame.from_address(id(self._tb.tb_frame)) # pylint: disable=no-member
                old.ob_refcnt -= 1
            if frame:
                assert isinstance(frame, types.FrameType)
                frame = TracebackProxy._Frame.from_address(id(frame)) # pylint: disable=no-member
                frame.ob_refcnt += 1
                self._obj.tb_frame = ctypes.pointer(frame)
            else:
                self._obj.tb_frame = ctypes.POINTER(TracebackProxy._Frame)()

        @property
        def tb_lasti(self):
            return self._tb.tb_lasti

        @tb_lasti.setter
        def tb_lasti(self, lasti):
            self._obj.tb_lasti = lasti

        @property
        def tb_lineno(self):
            return self._tb.tb_lineno

        @tb_lineno.setter
        def tb_lineno(self, lineno):
            self._obj.tb_lineno = lineno

        def __eq__(self, other):
            return self._tb == other._tb # pylint: disable=protected-access

        def __ne__(self, other):
            return self._tb != other._tb # pylint: disable=protected-access

        @staticmethod
        def create_traceback():
            try:
                1 / 0
            except:
                tb = sys.exc_info()[2]
            return tb


    def create_traceback_proxy(tb=None, frame_correction=0):
        """
        Builds a TracebackProxy object, using a given traceback, or current context if None.
        Returns a tuple with the first and the last tracebacks.
        Works only on CPython (Both Python 2 and 3 are supported)
        :param tb: traceback.traceback object to extract frames from
        :param frame_correction: Specifies the amount of frames to skip
        """
        assert frame_correction >= 0
        if isinstance(tb, types.TracebackType):
            for i in range(frame_correction + 1): # pylint: disable=unused-variable
                first = current = TracebackProxy(tb=tb)
                tb = tb.tb_next
            while tb:
                current.tb_next = TracebackProxy(tb=tb)
                tb = tb.tb_next
                current = current.tb_next
        else:
            frame_correction += 1 # Compensate this call frame
            frames = [frame_info[0] for frame_info in inspect.stack()[frame_correction:]]
            frames.reverse()
            first = current = TracebackProxy(frame=frames[0])
            for frame in frames[1:]:
                current.tb_next = TracebackProxy(frame=frame)
                current = current.tb_next

        return (first, current)



import functools
import re
import pytest
import warnings

import slash


def _warn(*args, **kwargs):
    return functools.partial(warnings.warn, *args, **kwargs)


def _catch(*args, **kwargs):
    return functools.partial(slash.ignore_warnings, *args, **kwargs)


class CustomWarning(UserWarning):
    pass


@pytest.mark.parametrize('emitter,catch,can_test_negative', [
    (_warn('hello'), _catch(message='hello'), True),
    (_warn('hello'), _catch(message=re.compile('^hello$')), True),
    (_warn('message', category=CustomWarning), _catch(category=CustomWarning), True),
    (_warn('message'), _catch(filename=__file__), False),
])
def test_ignore_warnings(request, emitter, catch, can_test_negative):

    request.addfinalizer(slash.clear_ignored_warnings)

    catch()

    with slash.Session() as session:
        emitter()
    assert list(session.warnings) == []

    if can_test_negative:
        with slash.Session() as session:
            warnings.warn('uncaught', category=DeprecationWarning)
        caught = list(session.warnings)
        assert len(caught) == 1
        assert caught[0].message == 'uncaught'



from __future__ import print_function

import sys
import traceback

from .. import hooks as trigger_hook
from ..conf import config
from ..ctx import context
from .. import exceptions
from .pattern_matching import Matcher

import warnings


def _debugger(debug_function_str, exc_info_transform=None, name=None):  # pragma: no cover
    module_name, function_name = debug_function_str.rsplit(".", 1)
    def debugger(exc_info):
        try:
            module = __import__(module_name, fromlist=[''])
        except ImportError:
            raise NotImplementedError() # pragma: no cover
        func = getattr(module, function_name)
        orig_exc_info = exc_info
        if exc_info_transform is not None:
            exc_info = exc_info_transform(exc_info)
        _notify_going_into_debugger(orig_exc_info)
        func(*exc_info)
    debugger.__name__ = debug_function_str
    debugger.__external_name__ = name
    return debugger

def _notify_going_into_debugger(exc_info):
    if context.session is not None:
        context.session.reporter.report_before_debugger(exc_info)
    else:
        print('\nException caught in debugger: {}'.format(traceback.format_exception_only(exc_info[0], exc_info[1])[0].strip()))

def _only_tb(exc_info):  # pragma: no cover
    return (exc_info[2],)

def _tb_type_value(exc_info):  # pragma: no cover
    return (exc_info[2], exc_info[0], exc_info[1])

_KNOWN_DEBUGGERS = [
    # order is important here!
    _debugger("pudb.post_mortem", _tb_type_value, name='pudb'),
    _debugger("ipdb.post_mortem", _only_tb, name='ipdb'),
    _debugger("pdb.post_mortem", _only_tb, name='pdb'),
    ]


def debug_if_needed(exc_info=None):

    if not config.root.debug.enabled:
        return
    if exc_info is None:
        exc_info = sys.exc_info()
    if exc_info[0] is None:
        return
    if isinstance(exc_info[1], context.session.get_skip_exception_types()) and not config.root.debug.debug_skips:
        return
    if isinstance(exc_info[1], (SystemExit,) + exceptions.INTERRUPTION_EXCEPTIONS):
        return

    exc_repr = repr(exc_info[1])
    matchers = [Matcher(s) for s in config.root.debug.filter_strings]
    if matchers and not all(matcher.matches(exc_repr) for matcher in matchers):
        return
    launch_debugger(exc_info)


def launch_debugger(exc_info):
    trigger_hook.entering_debugger(exc_info=exc_info) # pylint: disable=no-member

    debugger_name = config.root.debug.debugger

    debuggers = list(_KNOWN_DEBUGGERS)
    if debugger_name is not None:
        for index, debugger in enumerate(debuggers):
            if debugger_name == debugger.__external_name__:
                debuggers.insert(0, debuggers.pop(index))
                break
        else:
            warnings.warn('Specified debugger {!r} is not a known debugger name'.format(debugger_name))

    for debug_func in debuggers:
        try:
            debug_func(exc_info)
        except NotImplementedError:   # pragma: no cover
            if debug_func.__external_name__ == debugger_name:
                warnings.warn('Specified debugger {!r} is not available'.format(debugger_name))
            continue
        else:
            break
    else:
        raise NotImplementedError("No debug function available")  # pragma: no cover



import itertools
from collections import OrderedDict

from orderedset import OrderedSet

from ...exceptions import UnknownFixtures, InvalidFixtureScope, CyclicFixtureDependency

from ..._compat import ExitStack
from ...ctx import context
from .namespace import Namespace
from .parameters import iter_parametrization_fixtures
from .fixture_base import FixtureBase
from .utils import get_real_fixture_name_from_argument
from ..requirements import get_requirements

_fixture_id = itertools.count()



class Fixture(FixtureBase):

    def __init__(self, store, fixture_func):
        super(Fixture, self).__init__()
        self.fixture_func = fixture_func
        self.info = self.fixture_func.__slash_fixture__
        self.scope = self.info.scope
        self.namespace = Namespace(store, store.get_current_namespace())

    def is_parameter(self):
        return False

    def is_fixture(self):
        return True

    parametrization_ids = None

    def __repr__(self):
        return '<Function Fixture around {}>'.format(self.fixture_func)

    def is_override(self):
        parent = self.namespace.get_parent()

        while parent is not None:
            f = parent.get_fixture_by_name(self.info.name, default=None)
            if f is None:
                return False

            if f is not self:
                return True

            parent = parent.get_parent()
        return False

    def get_value(self, kwargs, active_fixture):
        if self.info.needs_this:
            assert 'this' not in kwargs
            kwargs['this'] = active_fixture
        with ExitStack() as stack:
            if context.session is not None:
                stack.enter_context(context.session.cleanups.default_scope_override(self.info.scope_name))
            return self.fixture_func(**kwargs)

    def get_requirements(self, store):
        fixture_requirements = get_requirements(self.fixture_func)
        required_fixtures = store.get_required_fixture_objects(self.fixture_func, self.namespace)
        while required_fixtures:
            fixture_requirements.extend(required_fixtures.pop().get_requirements(store))
        return fixture_requirements

    def _resolve(self, store):
        assert self.keyword_arguments is None
        assert self.parametrization_ids is None
        self.parametrization_ids = OrderedSet()
        keyword_arguments = OrderedDict()

        parametrized = set()

        for name, param in iter_parametrization_fixtures(self.fixture_func):
            store.register_fixture_id(param)
            parametrized.add(name)
            self.parametrization_ids.add(param.info.id)
            keyword_arguments[name] = param

        for param_name, arg in self.info.required_args.items():
            if param_name in parametrized:
                continue
            try:
                needed_fixture = self.namespace.get_fixture_by_name(get_real_fixture_name_from_argument(arg))

                if needed_fixture.scope < self.scope: # pylint: disable=no-member
                    raise InvalidFixtureScope('Fixture {} is dependent on {}, which has a smaller scope ({} > {})'.format(
                        self.info.name, param_name, self.scope, needed_fixture.scope)) # pylint: disable=no-member

                if needed_fixture is self:
                    raise CyclicFixtureDependency('Cyclic fixture dependency detected in {}: {} depends on itself'.format(
                        self.info.func.__code__.co_filename,
                        self.info.name))
                keyword_arguments[param_name] = needed_fixture
            except LookupError:
                raise UnknownFixtures(param_name)

        return keyword_arguments



import os
from collections import namedtuple

from . import pattern_matching

SuiteEntry = namedtuple('SuiteEntry', 'path, matcher, repeat')


def iter_suite_file_paths(suite_files):
    for filename in suite_files:

        dirname = os.path.abspath(os.path.dirname(filename))
        with open(filename) as suite_file:
            for path in suite_file:
                path = path.strip()
                if not path or path.startswith("#"):
                    continue

                suite_entry = _parse_path_filter_and_repeat(path)
                path = suite_entry.path

                if not os.path.isabs(path):
                    path = os.path.abspath(os.path.join(dirname, path))

                if not path.endswith('.py') and '.py:' not in path and not os.path.isdir(path):
                    for p, other_filter in iter_suite_file_paths([path]):
                        yield p, _and_matchers(suite_entry.matcher, other_filter)
                    continue

                for _ in range(suite_entry.repeat):
                    yield path, suite_entry.matcher


def _and_matchers(a, b):
    if a is None:
        return b
    if b is None:
        return a
    return pattern_matching.AndMatching([a, b])


def _parse_path_filter_and_repeat(line):
    if '#' not in line:
        return SuiteEntry(line, None, 1)

    line, remainders = line.split('#', 1)
    line = line.strip()
    remainders = remainders.split(',')

    matcher = None
    repeat = 1
    for remainder in remainders:
        remainder = remainder.strip()
        if remainder.startswith('filter:'):
            matcher = pattern_matching.Matcher(remainder.split(':', 1)[1])
        if remainder.startswith('repeat:'):
            repeat = int(remainder.split(':', 1)[1])
    return SuiteEntry(line, matcher, repeat)



from contextlib import contextmanager
from .utils.debug import debug_if_needed
from .utils.exception_mark import mark_exception, get_exception_mark
from .utils.traceback_proxy import create_traceback_proxy
from . import hooks as trigger_hook
from . import exceptions
from ._compat import PY2, PYPY
from .ctx import context as slash_context
from .conf import config

import functools
import threading
import logbook
try:
    import raven  # pylint: disable=F0401
except ImportError:
    raven = None
import sys

_logger = logbook.Logger(__name__)
NO_EXC_INFO = (None, None, None)


def update_current_result(exc_info):  # pylint: disable=unused-argument
    if slash_context.session is None:
        return
    if slash_context.test is not None:
        current_result = slash_context.session.results.get_result(slash_context.test)
    else:
        current_result = slash_context.session.results.global_result

    current_result.add_exception(exc_info)


def trigger_hooks_before_debugger(_):
    trigger_hook.exception_caught_before_debugger()  # pylint: disable=no-member


def trigger_hooks_after_debugger(_):
    trigger_hook.exception_caught_after_debugger()  # pylint: disable=no-member


_EXCEPTION_HANDLERS = [
    update_current_result,
    trigger_hooks_before_debugger,
    debug_if_needed,
    trigger_hooks_after_debugger,
]


class _IgnoredState(threading.local):
    ignored_exception_types = ()

_ignored_state = _IgnoredState()


class thread_ignore_exception_context(object):

    def __init__(self, exc_type):
        super(thread_ignore_exception_context, self).__init__()
        self._exc_type = exc_type
        self._prev = None

    def __enter__(self):
        self._prev = _ignored_state.ignored_exception_types
        _ignored_state.ignored_exception_types = list(_ignored_state.ignored_exception_types) + [self._exc_type]

    def __exit__(self, *_):
        _ignored_state.ignored_exception_types = self._prev
        self._prev = None


def handling_exceptions(fake_traceback=True, **kwargs):
    """Context manager handling exceptions that are raised within it

    :param passthrough_types: a tuple specifying exception types to avoid handling, raising them immediately onward
    :param swallow: causes this context to swallow exceptions
    :param swallow_types: causes the context to swallow exceptions of, or derived from, the specified types
    :param context: An optional string describing the operation being wrapped. This will be emitted to the logs to simplify readability

    .. note:: certain exceptions are never swallowed - most notably KeyboardInterrupt, SystemExit, and SkipTest
    """

    if not PYPY and fake_traceback:
        # Only in CPython we're able to fake the original, full traceback
        try:
            fake_tbs = create_traceback_proxy(frame_correction=2)
        except (KeyError, IndexError):
            _logger.warn("Could not extract full traceback for exceptions handling")
            fake_tbs = tuple()
    else:
        fake_tbs = tuple()
    swallow = kwargs.pop("swallow", False)
    swallow_types = kwargs.pop('swallow_types', ())
    if swallow:
        swallow_types = swallow_types + (Exception, )
    assert isinstance(swallow_types, (list, tuple)), 'swallow_types must be either a list or a tuple'
    passthrough_types = kwargs.pop('passthrough_types', ()) + tuple(_ignored_state.ignored_exception_types)
    return _HandlingException(fake_tbs, swallow_types, passthrough_types, kwargs)


class _HandledException(object):
    exception = None

class _HandlingException(object):

    def __init__(self, fake_tbs, swallow_types, passthrough_types, handling_kwargs):
        self._fake_traceback = fake_tbs
        self._kwargs = handling_kwargs
        self._passthrough_types = passthrough_types
        self._swallow_types = swallow_types
        self._handled = _HandledException()

    def __enter__(self):
        return self._handled

    def __exit__(self, *exc_info):
        if not exc_info or exc_info == NO_EXC_INFO:
            return
        exc_value = exc_info[1]

        if isinstance(exc_value, self._passthrough_types):
            return None

        if self._fake_traceback:
            (first_tb, last_tb) = self._fake_traceback
            (second_tb, _) = create_traceback_proxy(exc_info[2])
            last_tb.tb_next = second_tb
            exc_info = (exc_info[0], exc_info[1], first_tb._tb) # pylint: disable=protected-access
        handle_exception(exc_info, **self._kwargs)
        self._handled.exception = exc_info[1]
        skip_types = () if slash_context.session is None else slash_context.session.get_skip_exception_types()
        if isinstance(exc_value, skip_types) or isinstance(exc_value, exceptions.INTERRUPTION_EXCEPTIONS):
            return None
        if self._swallow_types and isinstance(exc_value, self._swallow_types):
            if PY2:
                sys.exc_clear()  # pylint: disable=no-member
            _logger.trace('Swallowing {!r}', exc_value)
            return True
        return None


def handle_exception(exc_info, context=None):
    """
    Call any handlers or debugging code before propagating an exception onwards.

    This makes sure that the exception can be handled as close as possible to its originating point.

    It also adds the exception to its correct place in the current result, be it a failure, an error or a skip

    """
    already_handled = is_exception_handled(exc_info[1])
    msg = "Handling exception"
    if context is not None:
        msg += " (Context: {0})"
    if already_handled:
        msg += " (already handled)"
    _logger.debug(msg, context, exc_info=exc_info if not already_handled else None)

    if not already_handled:
        mark_exception_handled(exc_info[1])
        for handler in _EXCEPTION_HANDLERS:
            handler(exc_info)


def mark_exception_handled(e):
    return mark_exception(e, "handled", True)


def is_exception_handled(e):
    """
    Checks if the exception ``e`` already passed through the exception handling logic
    """
    return bool(get_exception_mark(e, "handled", False))


@contextmanager
def get_exception_swallowing_context(report_to_sentry=True):
    """
    Returns a context under which all exceptions are swallowed (ignored)
    """
    try:
        yield
    except:
        if not get_exception_mark(sys.exc_info()[1], "swallow", True):
            raise
        if report_to_sentry:
            capture_sentry_exception()
        _logger.debug("Ignoring exception", exc_info=sys.exc_info())


def noswallow(exception):
    """
    Marks an exception to prevent swallowing by :func:`slash.exception_handling.get_exception_swallowing_context`,
    and returns it
    """
    mark_exception(exception, "swallow", False)
    return exception


def mark_exception_fatal(exception):
    """
    Causes this exception to halt the execution of the entire run.

    This is useful when detecting errors that need careful examination, thus preventing further tests from
    altering the test subject's state
    """
    mark_exception(exception, "fatal", True)
    return exception


def mark_exception_frame_correction(exception, correction=+1):
    current_correction = get_exception_frame_correction(exception)
    return mark_exception(exception, 'frame_correction', current_correction + correction)


def get_exception_frame_correction(exception):
    return get_exception_mark(exception, 'frame_correction', 0)


def is_exception_fatal(exception):
    return bool(get_exception_mark(exception, "fatal", False))


def inhibit_unhandled_exception_traceback(exception):
    """
    Causes this exception to inhibit console tracback
    """
    mark_exception(exception, "inhibit_console_tb", True)
    return exception


def should_inhibit_unhandled_exception_traceback(exception):
    return bool(get_exception_mark(exception, "inhibit_console_tb", False))


def disable_exception_swallowing(func_or_exception):
    """
    Marks an exception to prevent swallowing. Can also be used as a decorator around a function to mark all escaped
    exceptions
    """
    if isinstance(func_or_exception, BaseException):
        return noswallow(func_or_exception)

    @functools.wraps(func_or_exception)
    def func(*args, **kwargs):
        try:
            return func_or_exception(*args, **kwargs)
        except BaseException as e:
            disable_exception_swallowing(e)
            raise
    return func


def capture_sentry_exception():
    client = get_sentry_client()
    if client is not None:
        client.captureException()


def get_sentry_client():
    if raven is not None and config.root.sentry.dsn:
        return raven.Client(config.root.sentry.dsn)
    return None



import os
import logbook
from sqlalchemy import Column, DateTime, String, Integer
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.exc import OperationalError
from contextlib import contextmanager, closing
from datetime import datetime, timedelta
from slash.plugins import manager
from ast import literal_eval
from .utils.path import ensure_directory
from .conf import config

_DB_NAME = 'resume_state.db'
_MAX_DAYS_SAVED_SESSIONS = 10
_logger = logbook.Logger(__name__)
Base = declarative_base()
_session = sessionmaker()
_db_initialized = False


class ResumeState(Base):
    __tablename__ = 'resume_state'
    id = Column(Integer, primary_key=True)
    session_id = Column(String, nullable=False, index=True)
    file_name = Column(String, nullable=False, index=True)
    function_name = Column(String, nullable=False, index=True)
    variation = Column(String, index=True)


class SessionMetadata(Base):
    __tablename__ = 'session_metadata'
    session_id = Column(String, primary_key=True)
    src_folder = Column(String, nullable=False, index=True)
    created_at = Column(DateTime, nullable=False, index=True)


class ResumedTestData(object):

    def __init__(self, file_name, function_name, variation=None):
        self.file_name = file_name
        self.function_name = function_name
        self.variation = variation

    def __repr__(self):
        return '<ResumedTestData({!r}, {!r}, {!r})>'.format(self.file_name, self.function_name, self.variation)


def _init_db():
    resume_dir = os.path.expanduser(config.root.run.resume_state_path)
    ensure_directory(resume_dir)
    engine = create_engine('sqlite:///{}/{}'.format(resume_dir, _DB_NAME))
    _session.configure(bind=engine)
    Base.metadata.create_all(engine)


@contextmanager
def connecting_to_db():
    global _db_initialized  # pylint: disable=global-statement
    if not _db_initialized:
        _init_db()
        _db_initialized = True
    with closing(_session()) as new_session:
        yield new_session
        max_retries = 3
        for i in range(max_retries):
            try:
                new_session.commit() # pylint: disable=no-member
            except OperationalError:
                if i == max_retries - 1:
                    raise


def clean_old_entries():
    # pylint: disable=no-member
    with connecting_to_db() as conn:
        old_sessions_query = conn.query(SessionMetadata).filter \
            (SessionMetadata.created_at < datetime.now() - timedelta(days=_MAX_DAYS_SAVED_SESSIONS))
        old_sessions_ids = [session.session_id for session in old_sessions_query.all()]
        if old_sessions_ids:
            conn.query(ResumeState).filter(ResumeState.session_id.in_(old_sessions_ids)).delete(synchronize_session=False)
        old_sessions_query.delete(synchronize_session=False)


def save_resume_state(session_result, collected):
    session_metadata = SessionMetadata(
        session_id=session_result.session.id,
        src_folder=os.path.abspath(os.getcwd()),
        created_at=datetime.now())

    passed_tests_names = [result.test_metadata for result in session_result.iter_test_results() if result.is_success_finished()]
    collected_test_names = [test.__slash__ for test in collected]
    failed_metadata = list(set(collected_test_names) - set(passed_tests_names))

    tests_to_resume = []
    for metadata in failed_metadata:
        test_to_resume = ResumeState(session_id=session_result.session.id, file_name=metadata.file_path, function_name=metadata.function_name)
        if metadata.variation:
            test_to_resume.variation = str(metadata.variation.id)
        tests_to_resume.append(test_to_resume)

    with connecting_to_db() as conn:
        conn.add(session_metadata) # pylint: disable=no-member
        conn.add_all(tests_to_resume) # pylint: disable=no-member
    _logger.debug('Saved resume state to DB')


def get_last_resumeable_session_id():
    current_folder = os.path.abspath(os.getcwd())
    with connecting_to_db() as conn:
         # pylint: disable=no-member
        session_id = conn.query(SessionMetadata).filter(SessionMetadata.src_folder == current_folder) \
                                                .order_by(SessionMetadata.created_at.desc()).first()
        if not session_id:
            raise CannotResume("No sessions found for folder {}".format(current_folder))
        return session_id.session_id


def get_tests_from_previous_session(session_id, get_successful_tests=False):
    returned = []
    if get_successful_tests:
        return get_tests_from_remote_session(session_id, get_successful=True)
    with connecting_to_db() as conn:
         # pylint: disable=no-member
        session_metadata = conn.query(SessionMetadata).filter(SessionMetadata.session_id == session_id).first()
        if session_metadata:
            session_tests = conn.query(ResumeState).filter(ResumeState.session_id == session_id).all()
            if not session_tests:
                _logger.debug('Found session {} locally, but no failed tests exist', session_id)
                return returned
            for test in session_tests:
                new_entry = ResumedTestData(test.file_name, test.function_name)
                if test.variation:
                    new_entry.variation = literal_eval(test.variation)
                returned.append(new_entry)
    if not returned:
        _logger.debug('No local entry for session {}, searching remote session', session_id)
        returned = get_tests_from_remote_session(session_id)
    return returned


def get_tests_from_remote_session(session_id, get_successful=False):
    active_plugins = manager.get_active_plugins()
    backslash_plugin = active_plugins.get('backslash', None)
    if not backslash_plugin:
        raise CannotResume("Could not find backslash plugin")
    if not hasattr(backslash_plugin, 'is_session_exist') or not hasattr(backslash_plugin, 'get_tests_to_resume'):
        raise CannotResume("Backslash plugin doesn't support remote resuming")
    if not backslash_plugin.is_session_exist(session_id):
        raise CannotResume("Could not find resume data for session {}".format(session_id))
    remote_tests = backslash_plugin.get_tests_to_resume(session_id, get_successful)
    returned = [ResumedTestData(test.info['file_name'], test.info['name'], test.variation) for test in remote_tests]
    return returned


class CannotResume(Exception):
    pass



import sys
import os
import libtmux
import logbook
from ..exceptions import TmuxSessionNotExist, TmuxExecutableNotFound
from ..ctx import context

DEFAULT_SESSION_NAME = 'slash_session'
TMUX_EXECUTABLE_NAME = 'tmux'
MASTER_WINDOW_NAME = 'master'

_logger = logbook.Logger(__name__)

def which(program):
    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, _ = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file
    return None

def is_in_tmux():
    return os.environ.get('TMUX') is not None

def get_slash_tmux_session(session_name):
    try:
        tmux_server = libtmux.Server()
        return tmux_server.find_where({"session_name":session_name})
    except libtmux.exc.LibTmuxException:
        _logger.debug('No tmux server is running')
        return

def create_new_window(window_name, command):
    slash_session = get_slash_tmux_session(context.session.id)
    if not slash_session:
        raise TmuxSessionNotExist("Slash tmux session not found, can't create new window")
    return slash_session.new_window(attach=False, window_name=window_name, window_shell=command)

def create_new_pane(command):
    slash_session = get_slash_tmux_session(context.session.id)
    if not slash_session:
        raise TmuxSessionNotExist("Slash tmux session not found, can't create new window")
    new_pane = slash_session.attached_window.split_window(attach=False)
    new_pane.send_keys(command)
    return new_pane

def run_slash_in_tmux(command):
    tmux_session = get_slash_tmux_session(DEFAULT_SESSION_NAME)
    if tmux_session:
        tmux_session.set_option('remain-on-exit', True)
        libtmux.Server().switch_client(DEFAULT_SESSION_NAME)
        tmux_session.rename_session(context.session.id)
    else:
        path_to_tmux = which(TMUX_EXECUTABLE_NAME)
        if not path_to_tmux:
            _logger.error("Tmux executable not found")
            raise TmuxExecutableNotFound("Tmux executable not found")
        command = ' '.join([sys.executable, '-m', 'slash.frontend.main', 'run'] + command)
        tmux_args = [path_to_tmux, 'new-session', '-s', DEFAULT_SESSION_NAME, '-n', MASTER_WINDOW_NAME]
        if is_in_tmux():
            tmux_args.append('-Ad')
        tmux_args.append(command)
        os.execve(path_to_tmux, tmux_args, dict(os.environ))



import pytest
import operator

from slash._compat import reduce


def test_fixtures(suite, suite_test, defined_fixture):
    suite_test.depend_on_fixture(defined_fixture)
    suite.run()


def test_fixture_cleanup_at_end_of_suite(suite):
    fixture = suite.slashconf.add_fixture()
    suite[-1].depend_on_fixture(fixture)
    cleanup = fixture.add_cleanup()

    summary = suite.run()
    assert cleanup in summary.events


def test_fixture_cleanup_failure_fails_test(suite, suite_test, defined_fixture):
    suite_test.depend_on_fixture(defined_fixture)
    defined_fixture.add_cleanup(extra_code=['raise Exception()'])
    suite_test.expect_error()
    suite.run()


def test_fixture_parameters(suite, suite_test, defined_fixture):
    defined_fixture.add_parameter()
    suite_test.depend_on_fixture(defined_fixture)
    summary = suite.run()
    all_results = summary.get_all_results_for_test(suite_test)
    num_combinations = reduce(operator.mul, (len(p.values) for p in defined_fixture.get_parameters()))
    assert len(all_results) == num_combinations


def test_fixture_dependency_chain(suite, suite_test):
    fixture1 = suite.slashconf.add_fixture()
    fixture1.add_parameter()
    fixture2 = suite.slashconf.add_fixture()
    fixture2.add_parameter()
    fixture2.depend_on_fixture(fixture1)
    suite_test.depend_on_fixture(fixture2)
    suite.run()


def test_fixture_dependency_both_directly_and_indirectly(suite, suite_test):

    fixture1 = suite.slashconf.add_fixture()
    num_values1 = 2
    fixture1.add_parameter(num_values=num_values1)

    fixture2 = suite.slashconf.add_fixture()
    num_values2 = 3
    fixture2.add_parameter(num_values=num_values2)
    fixture2.depend_on_fixture(fixture1)

    suite_test.depend_on_fixture(fixture1)
    suite_test.depend_on_fixture(fixture2)

    summary = suite.run()
    results = summary.get_all_results_for_test(suite_test)
    assert len(results) == num_values1 * num_values2


def test_fixture_context(suite, suite_test):
    fixture1 = suite.slashconf.add_fixture()
    fixture1.append_line('assert this == slash.context.fixture')
    fixture2 = suite.slashconf.add_fixture()
    fixture2.append_line('assert this == slash.context.fixture')
    fixture2.depend_on_fixture(fixture1)
    suite_test.depend_on_fixture(fixture1)
    suite.run()

@pytest.mark.parametrize('where', [['before'], ['after'], ['before', 'after']])
def test_fixtures_in_before_after(suite, where):
    test_class = suite.files[-1].add_class()
    suite_test = test_class.add_method_test()

    fixture = suite.slashconf.add_fixture()
    fixture_event = fixture.add_event(name='fixture')

    assert len(suite_test.cls.tests) == 1

    before = suite_test.cls.add_before_method()
    evt1 = before.add_event(name='before')
    after = suite_test.cls.add_after_method()
    evt2 = after.add_event(name='after')

    for func in before, after:
        if func.name in where:
            func.depend_on_fixture(fixture)
            func.append_line('assert {} == {}'.format(fixture.name, fixture.get_value_string()))

    summary = suite.run()

    assert evt1 in summary.events
    assert evt2 in summary.events
    assert fixture_event in summary.events

def test_fixture_and_parameter(suite, suite_test, get_fixture_location):
    fixture = get_fixture_location(suite_test).add_fixture()
    suite_test.add_parameter()
    suite_test.depend_on_fixture(fixture)
    suite.run()


def test_fixture_which_name_startswith_test(suite):
    fixture = suite.slashconf.add_fixture(name='test_my_fixture')
    suite[-1].depend_on_fixture(fixture)
    for test in suite:
        test.expect_deselect()
    summary = suite.run(expect_session_errors=True)
    assert not summary.ok()
    for output in [summary.get_console_output(),
                   str(summary.session.results.global_result.get_errors()[0])]:
        assert 'Invalid fixture name' in output
